from unittest import TestCase

from basictracer import BasicTracer
from basictracer.recorder import InMemoryRecorder

from fan.context import Context
from fan.discovery import LocalDiscovery
from fan.remote import LocalEndpoint
from fan.tests import DummyService, NestedService


class FanTest(TestCase):
    def __init__(self, *args, **kwargs):
        super(FanTest, self).__init__(*args, **kwargs)

        self.recorder = InMemoryRecorder()
        discovery = LocalDiscovery()
        discovery.tracer = BasicTracer(self.recorder)

        for service in [DummyService, NestedService]:
            s = service()
            discovery.register(LocalEndpoint(s))

        self.context = Context(discovery)


class ServiceTest(FanTest):
    def test_01_simple(self):
        msg = 'test_message'
        with self.context:
            self.assertEqual(self.context.rpc.dummy.echo(msg), msg)

    def test_02_tree(self):
        msg = 'test_message'
        with self.context:
            self.assertEqual(self.context.rpc.nested.tree.dummy.echo(msg), msg)



import asyncio
import uuid
import aioredis

from fan.remote import RemoteEndpoint
from fan.contrib.aio.remote import AIOTransport, AIOQueueBasedTransport


class RedisStop(Exception):
    pass


class RedisTransport(AIOQueueBasedTransport, AIOTransport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import logging
        self.log = logging.getLogger(str(self))

    def new_connection(self):
        params = self.params
        return aioredis.create_redis((params.get('host', 'localhost'),
                                      params.get('port', 6379)),
                                     loop=self.loop)

    async def sub_prepare(self):
        self.sub = await self.new_connection()
        self.log.debug('Subscribe...')
        if self.remote:
            route = self.params['queue']
        else:
            route = self.back_route = str(uuid.uuid4())

        res = await self.sub.subscribe(route)
        self._read = asyncio.ensure_future(self.read_loop(res[0]))

    async def pub_prepare(self):
        self.pub = await self.new_connection()

    async def sub_stop(self):
        self.sub.close()

    async def pub_stop(self):
        self._read.cancel()
        self.pub.close()
        self.terminate(RedisStop())

    async def on_start(self):
        self.log.debug('Start redis')
        await super().on_start()

    async def rpc_inner_call(self, msg, resp):
        msg['back_route'] = self.back_route
        is_ok = await self.pub.publish_json(self.params['queue'], msg)
        # TODO: not clear what this code actually mean
        assert is_ok in (1, 2, 3), 'Not ok: {} => {}'.format(is_ok, self.endpoint)
        rep = await resp
        return rep['response']

    async def inner_read_message(self, chan):
        await chan.wait_message()
        return await chan.get_json()

    async def remote_send_response(self, msg, response):
        await self.pub.publish_json(msg['back_route'],
                                    {'context_headers': msg['context_headers'],
                                     'method': msg['method'],
                                     'response': response})


class RedisEndpoint(RemoteEndpoint):

    async def on_start(self):
        await self.transport.on_start()

    async def on_stop(self):
        await self.transport.on_stop()



import asyncio

from tipsi_tools.testing.aio import AIOTestCase
from fan.contrib.aio.lifecycle import (AIOSupervisor, AIOLifecycle, AIOSpec, MaxRestarts,
                                       State)


class Worker:
    def __init__(self):
        self.lifecycle = AIOLifecycle(self, 'on_start', 'terminate')
        self.started = False
        self.terminated = False

    async def on_start(self):
        self.started = True

    async def on_stop(self):
        await self.lifecycle.terminate()

    async def terminate(self, reason):
        self.started = False
        self.terminated = True


class DamagedWorker(Worker):
    async def new_on_start(self):
        self.started = True

    async def on_start(self):
        DamagedWorker.on_start = DamagedWorker.new_on_start
        raise Exception('damaged worker')


class BrokenWorkerError(Exception):
    pass


class BrokenWorker(Worker):
    def on_start(self):
        raise BrokenWorkerError


class TestSupervisor(AIOSupervisor):

    def __init__(self):
        self.specs = [AIOSpec(Worker)]
        super().__init__(self.specs)

    @property
    def state(self):
        return self.lifecycle.state


class AIOLifecycleTest(AIOTestCase):

    async def get_sup(self):
        if hasattr(self, '_s'):
            return self._s
        self._s = TestSupervisor()
        await self._s.lifecycle.on_start()
        return self._s

    async def test_01_simple_start(self):
        s = await self.get_sup()
        self.assertEqual(s.state, State.RUNNING)
        self.assertEqual(len(s.instances), 1)
        await s.lifecycle.terminate()
        self.assertEqual(s.state, State.TERMINATED)

    async def test_02_simple_add(self):
        s = await self.get_sup()
        worker = list(s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)

    async def test_03_worker_restart(self):
        s = await self.get_sup()
        worker = list(s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)
        await worker.lifecycle.terminate()
        self.assertEqual(worker.lifecycle.state, State.TERMINATED)

        worker = list(s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)

    async def test_04_damaged_restart(self):
        s = await self.get_sup()
        await s.add_child(AIOSpec(DamagedWorker))
        workers = list(s.instances)
        self.assertEqual(len(workers), 2)
        worker2 = workers[1]
        self.assertEqual(s.state, State.RUNNING)
        self.assertEqual(worker2.lifecycle.state, State.RUNNING)

        await s.lifecycle.terminate()
        self.assertEqual(s.lifecycle.state, State.TERMINATED)
        self.assertEqual(len(s.instances), 0)
        self.assertEqual(workers[0].lifecycle.state, State.TERMINATED)
        self.assertEqual(workers[1].lifecycle.state, State.TERMINATED)

    async def test_05_add_broken_worker(self):
        s = await self.get_sup()
        await self.assertRaises(MaxRestarts, s.add_child, AIOSpec(BrokenWorker))
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    async def test_06_broken_supervisor(self):
        s = AIOSupervisor([AIOSpec(BrokenWorker)])
        await self.assertRaises(MaxRestarts, s.lifecycle.on_start)
        self.assertEqual(list(s.instances), [])
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    async def test_07_supervisor_of_supervisor(self):
        sup_spec = AIOSpec(AIOSupervisor, [AIOSpec(BrokenWorker)])
        s = AIOSupervisor([sup_spec])
        await self.assertRaises(MaxRestarts, s.lifecycle.on_start)
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    async def test_08_nested_add(self):
        sup_spec = AIOSpec(AIOSupervisor, [AIOSpec(DamagedWorker)])
        s = AIOSupervisor([sup_spec])
        await s.lifecycle.on_start()
        self.assertEqual(s.lifecycle.state, State.RUNNING)

        s2 = list(s.instances)[0]
        self.assertEqual(s2.lifecycle.state, State.RUNNING)

        await self.assertRaises(MaxRestarts, s2.add_child, AIOSpec(BrokenWorker))
        self.assertEqual(s2.lifecycle.state, State.TERMINATED)
        self.assertEqual(s.lifecycle.state, State.TERMINATED)


class LaggyWorker:
    def __init__(self):
        self.lifecycle = AIOLifecycle(self, 'on_start', 'terminate')
        self.started = False
        self.terminated = False

    async def on_start(self):
        self.started = True
        asyncio.ensure_future(self.on_stop())

    async def on_stop(self):
        await self.lifecycle.terminate()

    async def terminate(self, reason):
        self.started = False
        self.terminated = True


class Linked(AIOLifecycle):
    def __init__(self):
        self.lifecycle = AIOLifecycle(self, notify_method='notify')
        self.notified = False

    def notify(self, reason, obj):
        self.notified = True

class AIOOnlyTest(AIOTestCase):
    async def test_01_die_after_start(self):
        s = AIOSupervisor([AIOSpec(LaggyWorker)])
        await s.lifecycle.on_start()
        self.assertEqual(s.lifecycle.state, State.RUNNING)
        await asyncio.sleep(0.01)
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    async def test_02_restart(self):
        s = AIOSupervisor([AIOSpec(Worker)])
        await s.lifecycle.on_start()
        self.assertEqual(s.lifecycle.state, State.RUNNING)
        w = list(s.instances)[0]
        l = Linked()
        l.lifecycle.watch(w)
        self.assertEqual(l.notified, False)
        self.ensure_future(w.lifecycle.terminate())
        await asyncio.sleep(0.001)
        self.assertEqual(l.notified, True)
        self.assertEqual(w.lifecycle.state, State.TERMINATED)

        self.assertEqual(s.lifecycle.state, State.RUNNING)



import time
import weakref
import logging
from enum import IntEnum
from collections import OrderedDict


class LifecycleReason(Exception):
    pass


class SupervisorStop(LifecycleReason):
    pass


class Terminated(LifecycleReason):
    pass


class MaxRestarts(LifecycleReason):
    pass


class RestartType(IntEnum):
    one_for_one = 1
    rest_for_one = 2
    one_for_all = 3


class State(IntEnum):
    INITIALIZED = 0
    STARTING = 1
    RUNNING = 2
    STOPPING = 3
    STOPPED = 4
    TERMINATING = 5
    TERMINATED = 6


class RestartStrategy:
    def __init__(self, restart_type=RestartType.one_for_one, max_r=10, max_t=1):
        self.restart_type = RestartType
        self.max_r = max_r
        self.max_t = max_t


def terminate_on_error(fun):
    def wrapped(self, *args, **kwargs):
        try:
            res = fun(self, *args, **kwargs)
            return res
        except Exception as e:
            self.log.debug('Catch exception: {!r}'.format(e))
            try:
                if hasattr(self, 'lifecycle'):
                    self.lifecycle.terminate(e)
                else:
                    self.terminate(e)
            except Exception as term_e:
                self.log.exception('during terminate')
            raise e
    return wrapped


class Spec:
    def __init__(self, cls, *args, strategy=RestartStrategy(), **kwargs):
        self.cls = cls
        self.args = args
        self.strategy = strategy
        self.kwargs = kwargs
        self.last_create = 0
        self.restarts = 0

    def get_obj(self):
        curr = time.time()
        if curr - self.last_create < self.strategy.max_t:
            self.restarts += 1
            if self.restarts >= self.strategy.max_r:
                raise MaxRestarts
        else:
            self.restarts = 0
        self.last_create = curr
        return self.cls(*self.args, *self.kwargs)

    def run(self, obj):
        obj.lifecycle.on_start()
        return obj


class Lifecycle:
    '''
    We assume that you will set self.lifecycle = Lifecycle()
    '''
    log = logging.getLogger('Lifecycle')

    def __init__(self, obj, start_method='', terminate_method='', notify_method=''):
        self.obj = obj
        self.start_method = start_method
        self.terminate_method = terminate_method
        self.notify_method = notify_method
        self.links = weakref.WeakSet()
        self.watches = weakref.WeakSet()
        self.state = State.INITIALIZED

    @terminate_on_error
    def on_start(self):
        assert self.state == State.INITIALIZED, self.state
        self.state = State.STARTING
        if self.start_method:
            getattr(self.obj, self.start_method)()
        self.state = State.RUNNING

    def safe_call(self, fun, *args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except Exception as e:
            self.log.info('during call {} [{} {}]'.format(fun, args, kwargs))

    def terminate(self, reason=Terminated):
        if self.state >= State.STOPPING:
            return
        self.state = State.STOPPING
        self.log.info('Terminate reason: {!r} {}'.format(reason, self.obj))
        for link in self.links:
            self.safe_call(link.terminate, reason, self.obj)
        for watch in self.watches:
            self.safe_call(watch.notify, reason, self.obj)
        self.state = State.STOPPED
        if self.terminate_method:
            try:
                self.state = State.TERMINATING
                self.safe_call(getattr(self.obj, self.terminate_method), reason)
            except Exception as e:
                self.log.exception('during terminate: {}.{}'.format(self.obj,
                                                                    self.terminate_method))
        self.state = State.TERMINATED

    @terminate_on_error
    def notify(self, reason, obj):
        getattr(self.obj, self.notify_method)(reason, obj)

    def link(self, obj):
        assert self.state <= State.RUNNING
        obj.lifecycle.links.add(self)

    def watch(self, obj):
        assert self.state <= State.RUNNING
        obj.lifecycle.watches.add(self)


class Supervisor:
    log = logging.getLogger('Supervisor')

    def __init__(self, specs=[]):
        assert isinstance(specs, list), 'incompatible specs: {!r}'.format(specs)
        self.specs = specs
        self.instances = OrderedDict()
        self.lifecycle = Lifecycle(self, 'on_start', 'terminate', 'notify')

    @terminate_on_error
    def notify(self, reason, obj):
        if self.lifecycle.state > State.RUNNING:
            return
        try:
            spec = self.instances.pop(obj)
        except KeyError as e:
            self.log.exception('Get notify without from obj not in self.instances')
            raise e
        self.start_child(spec)

    def add_child(self, spec):
        assert self.lifecycle.state == State.RUNNING, \
          'Supervisor should be started before add child {}'.format(spec)
        self.specs.append(spec)
        self.start_child(spec)

    @terminate_on_error
    def start_child(self, spec):
        while True:
            try:
                child = spec.get_obj()
                spec.run(child)
                self.lifecycle.watch(child)
                self.instances[child] = spec
                return child
            except MaxRestarts as e:
                raise e
            except Exception:
                self.log.info('Restart worker: {}'.format(spec))

    def on_start(self):
        self.log.debug('Start supervisor')
        assert self.lifecycle.state == State.STARTING, "don't call on_start directly"
        for spec in self.specs:
            self.start_child(spec)

    def terminate(self, reason):
        assert self.lifecycle.state == State.TERMINATING, "don't call terminate directly"
        self.log.info('Supervisor terminate reason: {}'.format(reason))
        while True:
            try:
                obj, spec = self.instances.popitem()
                obj.lifecycle.terminate(reason)
            except KeyError:
                break



import os

from fan.contrib.amqp import AMQPEndpoint, AMQPTransport
from fan.contrib.aio.tests import AIOEndpointCase


class AMQPCase(AIOEndpointCase):
    endpoint_class = AMQPEndpoint
    transport_class = AMQPTransport
    endpoint_params = {'host': os.environ.get('AMQP_HOST', 'rabbitmq'),
                       'queue': 'dummy',
                       'exchange': 'tipsi',
                       'exchange_type': 'topic',
                       'routing_key': 'dummy',
                       'transport': 'amqp'}

    async def test_call(self):
        await self._test_remote_call()

    async def test_remote_register(self):
        await self._test_remote_register()



from fan.contrib.django import VARS
from tipsi_tools.logging import JSFormatter


class SpanFormatter(JSFormatter):
    def process_log_record(self, rec):
        rec.update(VARS)
        return super().process_log_record(rec)



import asyncio
import asynqp
import re
import time
from types import CoroutineType

from basictracer.context import SpanContext

from fan.context import Context
from fan.remote import RemoteEndpoint
from fan.contrib.aio.remote import AIOTransport, AIOQueueBasedTransport


asynqp.queue.VALID_QUEUE_NAME_RE = re.compile('')


class AMQPTransport(AIOQueueBasedTransport, AIOTransport):
    async def ensure_connection(self):
        if hasattr(self, 'conn'):
            return
        params = self.params
        while True:
            try:
                self.conn = await asynqp.connect(params.get('host', 'localhost'),
                                                 params.get('port', 5672),
                                                 username=params.get('user', 'guest'),
                                                 password=params.get('password', 'guest'))
                break
            except OSError:
                time.sleep(0.2)

    async def sub_prepare(self):
        params = self.params
        await self.ensure_connection()
        self.sub = await self.conn.open_channel()
        self.routing_key = params['routing_key']
        self.default_exchange = await self.sub.declare_exchange('', 'direct')
        self.exchange = await self.sub.declare_exchange(params.get('exchange', ''),
                                                        params.get('exchange_type', 'direct'))
        self.log.debug('Subscribe...')
        if self.remote:
            queue_name = params['queue']
            queue = await self.sub.declare_queue(queue_name)
            await queue.bind(self.exchange, self.routing_key)
        else:
            queue_name = 'amq.rabbitmq.reply-to'
            queue = await self.sub.declare_queue(queue_name, nowait=True)
        # self.loop.create_task(self.read_loop(queue))
        self.queue = queue
        await queue.consume(self.deliver, no_ack=not self.remote)

    async def pub_prepare(self):
        self.pub = await self.conn.open_channel()

    async def sub_stop(self):
        await self.sub.close()
        await self.conn.close()

    async def pub_stop(self):
        await self.pub.close()

    async def on_start(self):
        self.log.info('Start amqp')
        await super().on_start()

    async def rpc_inner_call(self, msg, resp):
        ctx = msg['context_headers']
        for k, v in ctx.items():
            ctx[k] = str(v)
        body = msg
        span_id = str(ctx['span_id'])
        amqp_msg = asynqp.Message(body,
                                  headers=ctx,
                                  reply_to='amq.rabbitmq.reply-to',
                                  correlation_id=span_id)
        self.log.debug('Publish message: {} {}'.format(self.exchange, self.routing_key))
        self.exchange.publish(amqp_msg, self.routing_key)
        rep = await resp
        return rep

    def deliver(self, msg):
        self.log.debug('DELIVERED: {}'.format(msg))
        self.loop.create_task(self.read_loop(msg))

    async def read_loop(self, raw_msg):
        try:
            self.log.debug('Got message: {} Q[{}]'.format(raw_msg, self.queue.name))
            msg = raw_msg.json()
            ctx_headers = msg['context_headers']
            if self.remote:
                parent_ctx = SpanContext(**ctx_headers)
                method = msg['method']
                ctx = Context(self.discovery, self.endpoint.service, parent_ctx, method)
                self.log.debug('CTX: {} {}'.format(ctx.span.context.trace_id, raw_msg.correlation_id))
                args = msg.get('args', ())
                kwargs = msg.get('kwargs', {})
                hc = self.handle_call(method, ctx, *args, **kwargs)
                if isinstance(hc, CoroutineType):
                    resp = await hc
                else:
                    resp = hc
                self.log.debug('Send resp ==> : {}'.format(resp))
                resp = asynqp.Message({'context_headers': msg['context_headers'],
                                       'method': msg['method'],
                                       'response': resp},
                                      correlation_id=raw_msg.correlation_id)
                self.default_exchange.publish(resp, raw_msg.reply_to, mandatory=False)
                raw_msg.ack()
            else:
                self.proxy_send_response(msg)
        except Exception as e:
            self.terminate(e)


class AMQPEndpoint(RemoteEndpoint):
    async def on_start(self):
        await self.transport.on_start()

    async def on_stop(self):
        await self.transport.on_stop()



import asyncio
from fan.process import Process


class AIOProcess(Process):
    async def run(self):
        pass

    def start(self):
        self.loop = asyncio.get_event_loop()
        self.init_process()
        self.loop.run_forever()



import asyncio
import json
import logging

from aiozk import ZKClient, exc

from fan.discovery import CompositeDiscovery, RemoteDiscovery
from fan.contrib.aio.remote import AIOProxyEndpoint


class ZKDiscovery(RemoteDiscovery):
    log = logging.getLogger('fan.aio.ZKDiscovery')

    def __init__(self, zk_path, chroot=None):
        super().__init__()
        self.zk = ZKClient(zk_path, chroot)

    async def create(self, path, **kwargs):
        try:
            return await self.zk.create(path, **kwargs)
        except exc.NodeExists:
            pass

    async def recursive_create(self, path, data) -> str:
        if not isinstance(path, list):
            path = path.split('/')
        curr = '/'
        await self.create(curr)
        for sub in path:
            curr += '/{}'.format(sub)
            await self.create(curr, container=True)
        curr += '/config_'
        if not (isinstance(data, str) or isinstance(data, bytes)):
            data = json.dumps(data).encode()
        out = await self.create(curr, data=data, ephemeral=True, sequential=True)
        return out, data

    async def on_start(self):
        await self.zk.start()
        if not await self.zk.exists('/endpoints'):
            await self.zk.create('/endpoints')
        asyncio.ensure_future(self._check_zk())
        self.data_watcher = self.zk.recipes.DataWatcher()
        self.data_watcher.set_client(self.zk)

    async def _check_zk(self):
        try:
            while True:
                await self.zk.session.ensure_safe_state()
                self.log.debug('ZK session is ok. looping...')
                await asyncio.sleep(5)
        except Exception as e:
            self.log.exception('Session closed')
            asyncio.get_event_loop().stop()

    async def register(self, endpoint) -> str:
        name = endpoint.name
        path = name.split('.')
        path.append(endpoint.version)
        barrier_path = '/'.join(path + ['barrier'])
        return await self.register_raw(path, endpoint.config, barrier_path)

    async def register_raw(self, path, data, barrier_path=None) -> str:
        if barrier_path:
            barrier = self.zk.recipes.Barrier(barrier_path)
            await barrier.wait()
        return await self.recursive_create(path, data)

    async def find_endpoint(self, service_name, version_filter):
        path = ['/endpoints'] + service_name.split('.')
        path = '/'.join(path)

        if not await self.zk.exists(path):
            return
        childs = await self.zk.get_children(path)
        if len(childs) == 0:
            return

        version = sorted(childs)[-1]
        vpath = '{}/{}'.format(path, version)
        configs = await self.zk.get_children(vpath)
        # keep only config_ paths
        configs = filter(lambda x: x.startswith('config_'), configs)
        return await self.create_endpoint(service_name, vpath, configs)

    async def create_endpoint(self, name, path, configs):
        # TODO: pass all configs
        config = sorted(configs)[-1]
        dpath = '{}/{}'.format(path, config)
        data = await self.zk.get_data(dpath)
        print('GET DATA: {!r}'.format(data))
        params = json.loads(data)
        return AIOProxyEndpoint(self, name, params)

    def watch(self, path, callback):
        self.data_watcher.add_callback(path, callback)
        pass

    def unwatch(self, path, callback):
        self.data_watcher.remove_callback(path, callback)


class AIOCompositeDiscovery(CompositeDiscovery):
    def create_proxy(self, name, proxy_cfg):
        return AIOProxyEndpoint(self, name, proxy_cfg)



'''
Shortlived sync helpers. Primary target is creating short-lived fast context with get_context
'''
from functools import wraps
import json
import logging
import os
import socket
import time

import requests
from basictracer import BasicTracer
from basictracer.recorder import InMemoryRecorder
from py_zipkin.zipkin import ZipkinAttrs
from py_zipkin.thrift import (annotation_list_builder, create_endpoint,
                              binary_annotation_list_builder,
                              create_span, thrift_objs_in_bytes)

from fan.context import Context
from fan.contrib.kazoo.discovery import KazooDiscovery
from fan.exceptions import DiscoveryConnectionError
from fan.transport import HTTPTransport, HTTPPropagator, DjangoPropagator


discovery = None
tracer = None
ZIPKIN = os.environ.get('ZIPKIN')


def http_transport(encoded_span):
    # copypasted from: https://github.com/Yelp/py_zipkin#transport
    # The collector expects a thrift-encoded list of spans. Instead of
    # decoding and re-encoding the already thrift-encoded message, we can just
    # add header bytes that specify that what follows is a list of length 1.
    requests.post(
        'http://{}/api/v1/spans'.format(ZIPKIN),
        data=encoded_span,
        headers={'Content-Type': 'application/x-thrift'},
    )


MY_IP = socket.gethostbyname(socket.gethostname())
EP = None
log = logging.getLogger('fan.sync')


def logger_log_span(span_id, parent_span_id, trace_id, span_name, annotations,
                    binary_annotations, **kwargs):
    log.info('Log span: {}'.format(locals()))


def zipkin_log_span(span_id, parent_span_id, trace_id, span_name, annotations,
                    binary_annotations, timestamp_s, duration_s, **kwargs):
    span = create_span(span_id, parent_span_id, trace_id, span_name,
                       annotations, binary_annotations,
                       timestamp_s, duration_s)
    http_transport(thrift_objs_in_bytes([span]))

    params = {'timestamp_s': timestamp_s,
              'duration_s': duration_s,
              **kwargs}
    logger_log_span(span_id, parent_span_id, trace_id, span_name, annotations,
                    binary_annotations, **params)


class FanRecorder(InMemoryRecorder):
    def __init__(self, name, log_span):
        super().__init__()
        self.name = name
        self.log_span = log_span

    def record_span(self, span):
        global EP
        if not EP:
            EP = create_endpoint(port=80, service_name=self.name, host=MY_IP)

        ctx = span.context
        timing = {'ss': span.start_time,
                  'sr': span.start_time+span.duration}
        annotations = annotation_list_builder(timing, EP)
        attrs = ZipkinAttrs(trace_id=hex(ctx.trace_id)[2:],
                            span_id=hex(ctx.span_id)[2:],
                            parent_span_id=span.parent_id and hex(span.parent_id)[2:],
                            flags=0,
                            is_sampled=True)
        params = {
            'zipkin_attrs': attrs,
            'trace_id': hex(ctx.trace_id)[2:],
            'span_id': hex(ctx.span_id)[2:],
            'parent_span_id': span.parent_id and hex(span.parent_id)[2:],
            'service_name': self.name,
            'span_name': span.operation_name or 'no_name',
            'annotations': annotations,
            'binary_annotations': binary_annotation_list_builder({**span.context.baggage,
                                                                  **span.tags}, EP),
            'transport_handler': http_transport,
            'timestamp_s': span.start_time,
            'duration_s': span.duration,
        }

        self.log_span(**params)
        return super().record_span(span)


def get_tracer(name=None):
    global tracer
    if tracer:
        return tracer
    if not name:
        name = 'no_name'

    if ZIPKIN:
        log_span = zipkin_log_span
    else:
        log_span = logger_log_span
    recorder = FanRecorder(name, log_span)
    tracer = BasicTracer(recorder)
    return tracer


def cache_discovery(fun):
    @wraps(fun)
    def wrapped(*args, **kwargs):
        global discovery
        if discovery:
            return discovery
        discovery = fun(*args, **kwargs)
        return discovery
    return wrapped


@cache_discovery
def get_discovery(is_django=False, name=None):
    discovery = KazooDiscovery(os.environ.get('ZK_HOST', 'zk'), os.environ.get('ZK_CHROOT', '/'))
    discovery.transport_classes = {
        'http': HTTPTransport,
    }

    if not name and is_django:
        name = 'django'

    discovery.tracer = get_tracer(name)
    if is_django:
        discovery.tracer.register_propagator('http', DjangoPropagator())
    else:
        discovery.tracer.register_propagator('http', HTTPPropagator())
    return discovery


def get_context(name=None, service_name=None):
    if not service_name and name:
        service_name = name
    return Context(get_discovery(is_django=False, name=service_name), name=name)



import asyncio
import logging

from basictracer import BasicTracer
from basictracer.recorder import InMemoryRecorder
from tipsi_tools.testing.aio import AIOTestCase

from fan.context import Context
from fan.contrib.aio.remote import AIOProxyEndpoint, AIOTransport
from fan.remote import RemoteEndpoint

from fan.discovery import SimpleDictDiscovery, LocalDiscovery
from fan.contrib.aio.discovery import AIOCompositeDiscovery
from fan.service import Service, endpoint
from fan.tests import TEST_TIMEOUT


class AIODummyTransport(AIOTransport):
    requests = {}

    async def on_start(self):
        if self.remote:
            self.channel = asyncio.Queue()
            self.requests[self.params['id']] = self.channel
        else:
            self.channel = asyncio.Queue()
        self._read = asyncio.ensure_future(self.read_loop())

    async def on_stop(self):
        self._read.cancel()

    async def rpc_inner_call(self, msg, future):
        msg['response'] = self.channel
        await self.requests[self.params['id']].put(msg)
        return await future


class AIODummyEndpoint(RemoteEndpoint):
    async def on_start(self):
        await self.transport.on_start()

    async def on_stop(self):
        await self.transport.on_stop()


class TestLocalDiscovery(LocalDiscovery):
    transports = {'aiodummy': AIODummyTransport}

    def get_transport_class(self, name):
        return self.transports[name]


class DummyService(Service):
    name = 'dummy'

    @endpoint
    def ping(self, ctx):
        self.log.debug('Call ping endpoint {}'.format(ctx.span.context.trace_id))
        return 'pong'


class AIOEndpointCase(AIOTestCase):
    endpoint_class = AIODummyEndpoint
    transport_class = AIODummyTransport
    endpoint_params = {'id': 'test', 'transport': 'aiotest'}

    def __init__(self, *args):
        super().__init__(*args)
        self.log = logging.getLogger(self.__class__.__name__)

    def get_local_discovery(self):
        ld = TestLocalDiscovery()
        TestLocalDiscovery.transports = {self.endpoint_params['transport']: self.transport_class}
        if not hasattr(self, 'ctx_local'):
            self.ctx_local = []
        self.ctx_local.append(ld)
        return ld

    async def setUp(self):
        self.recorder = InMemoryRecorder()
        self.dict_discovery = SimpleDictDiscovery({})
        self.discovery = AIOCompositeDiscovery(self.get_local_discovery(),
                                               self.dict_discovery)
        self.discovery.tracer = BasicTracer(self.recorder)
        self.svc = DummyService()
        await self.ensure_future(self.svc.on_start())

    async def _test_remote_register(self):
        ep = self.endpoint_class(self.discovery, self.svc, self.endpoint_params)
        await asyncio.wait_for(ep.on_start(), TEST_TIMEOUT)
        self.discovery.register(ep)

        l = self.discovery.local
        r = self.discovery.remote
        assert l.cached_endpoints[('dummy',)] == ep
        assert r.data['dummy'] == self.endpoint_params

    @property
    def ctx(self):
        ld = TestLocalDiscovery()
        self.ctx_local.append(ld)
        discovery = AIOCompositeDiscovery(self.get_local_discovery(),
                                          self.dict_discovery)
        discovery.tracer = BasicTracer(self.recorder)
        return Context(discovery, self.svc)

    async def _test_remote_call(self):
        params = self.endpoint_params
        ep = self.endpoint_class(self.discovery, self.svc, params)
        await asyncio.wait_for(ep.on_start(), TEST_TIMEOUT)
        self.discovery.register(ep)

        with self.ctx as ctx:
            res = await asyncio.wait_for(ctx.rpc.dummy.ping(), TEST_TIMEOUT)
        self.assertEqual(res, 'pong')
        self.assertEquals(len(self.recorder.get_spans()), 2)
        spans = [(x.context.span_id, x.context.trace_id) for x in self.recorder.get_spans()]
        self.log.debug('SPANS: {}'.format(spans))

    async def tearDown(self):
        for ld in self.ctx_local:
            for v in ld.cached_endpoints.values():
                await v.on_stop()



import logging

from django.conf import settings
from django.http import HttpResponse

from fan.context import Context
from fan.exceptions import RPCHttpError
from fan.sync import get_discovery

VARS = {}


def update_vars(ctx, request):
    span = ctx.span
    _ctx = ctx.span.context
    VARS['TRACE_ID'] = hex(_ctx.trace_id)[2:]
    VARS['SPAN_ID'] = hex(_ctx.span_id)[2:]
    VARS['PARENT_SPAN_ID'] = span.parent_id and hex(span.parent_id)[2:]
    # HTTP_INSTALLATION_ID
    # HTTP_SESSION_ID
    for k, v in _ctx.baggage.items():
        VARS[k.upper()] = v


class FanMiddleware(object):
    '''
    Enable tracing handling
    Adds `request.ctx`
    '''
    log = logging.getLogger('FanMiddleware')

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        '''
        We're performing initialization when request.ctx is called
        So set property here
        '''
        if hasattr(request, 'ctx'):
            return self.get_response(request)
        discovery = get_discovery(is_django=True, name=getattr(settings, 'FAN_SERVICE', None))
        tracer = discovery.tracer

        span_context = tracer.extract('http', request.META)
        name = '{} {}'.format(request.method, request.path)
        if span_context:
            ctx = Context(discovery, None, span_context, name)
        else:
            # TODO: should be configurable
            baggage = {}
            for k in ['HTTP_INSTALLATION_ID', 'HTTP_SESSION_ID']:
                if k in request.META:
                    baggage[k.strip('HTTP_')] = request.META[k]
            ctx = Context(discovery, None, None, name, baggage)
        request.ctx = ctx
        with ctx:
            update_vars(ctx, request)
            return self.get_response(request)

    def process_exception(self, request, exception):
        if isinstance(exception, RPCHttpError):
            resp = exception.response
            return HttpResponse(
                content=resp.content,
                status=resp.status_code,
                content_type=resp.headers['Content-Type']
            )



class RPCException(Exception):
    pass


class RPCHttpError(RPCException):
    def __init__(self, response):
        self.response = response
        super().__init__(self.response)

    def __str__(self):
        return '<RPCHttpError {} {}'.format(self.response.status_code,
                                            self.response.content.decode('utf8'))


class DiscoveryConnectionError(Exception):
    pass



import logging
import requests
import json

from basictracer.propagator import Propagator
from basictracer.context import SpanContext

from fan.remote import Transport
from fan.exceptions import RPCHttpError


def hex_string(i):
    return '{:x}'.format(i)


def from_hex_string(s):
    return int(s, 16)


class HTTPPropagator(Propagator):
    log = logging.getLogger('fan.HTTPPropagator')
    mapping = {
        # name: (key_name, dump, extract)
        'span_id': ('ot-span-id', hex_string, from_hex_string),
        'trace_id': ('ot-trace-id', hex_string, from_hex_string),
        'baggage': ('ot-baggage', json.dumps, json.loads),
        'sampled': ('ot-sampled', json.dumps, json.loads),
    }

    def inject(self, span_context, carrier):
        for k, v in self.mapping.items():
            name, pack, _ = v
            carrier[name] = pack(getattr(span_context, k))
        return carrier

    def extract(self, carrier):
        # TODO: support ot-debug => override sampled to true
        # generate custom label
        try:
            kwargs = {'baggage': {}}
            for k, v in self.mapping.items():
                name, _, extract = v
                kwargs[k] = extract(carrier[name])
            return SpanContext(**kwargs)
        except Exception:
            pass


class DjangoPropagator(HTTPPropagator):
    mapping = {
        'span_id': ('HTTP_OT_SPAN_ID', hex_string, from_hex_string),
        'trace_id': ('HTTP_OT_TRACE_ID', hex_string, from_hex_string),
        'baggage': ('HTTP_OT_BAGGAGE', json.dumps, json.loads),
        # 'baggage': 'ot-baggage',
        'sampled': ('HTTP_OT_SAMPLED', json.dumps, json.loads),
    }

    def extract(self, carrier):
        self.log.debug('Run extract: {}'.format(carrier))
        try:
            kwargs = {'baggage': {}}
            for k, v in self.mapping.items():
                name, _, extract = v
                key = name if name in carrier else 'HTTP_{}'.format(name)
                # crash when no HTTP_OT_SPAN_ID => return None
                kwargs[k] = extract(carrier[key])
            return SpanContext(**kwargs)
        except Exception:
            pass


class HTTPTransport(Transport):
    log = logging.getLogger('HTTPTransport')

    def __init__(self, discovery, endpoint, params):
        super().__init__(discovery, endpoint, params)
        self.base_url = '{transport}://{host}:{port}'.format(**params)
        self.methods = {}
        for method in params['methods']:
            self.methods[method['name']] = method

    def get_headers(self, ctx):
        hdrs = {}
        tracer = ctx.discovery.tracer
        tracer.inject(ctx.span.context, 'http', hdrs)
        return hdrs

    def prepare_get_params(self, params):
        out = {}
        for k, v in params.items():
            if isinstance(v, list):
                out[k] = ','.join([str(x) for x in v])
            else:
                out[k] = v
        return out

    def rpc_call(self, method_name, ctx, **kwargs):
        assert method_name in self.methods, 'Cannot find endpoint'
        method = self.methods[method_name]
        url = ''.join([self.base_url, method['url']])
        if '{' in url:
            url = url.format(**kwargs)
        m = method.get('method', 'get').lower()
        req = getattr(requests, m)

        if m in ('get', 'delete'):
            kw = {'params': self.prepare_get_params(kwargs)}
        else:
            kw = {'json': kwargs}
        self.log.debug('Url: {} Params: {} Func: {}'.format(url, kw, req))
        kw['headers'] = self.get_headers(ctx)
        resp = req(url, **kw)
        if resp.status_code in (200, 201):
            ret = resp.json()
        elif resp.status_code in (204,):
            ret = True
        else:
            # TODO: howto return error
            self.log.error('Resp: {} : {}'.format(resp.status_code, resp))
            raise RPCHttpError(resp)
        return ret



from basictracer import BasicTracer
from fan.context import Context
from typing import List, Dict, Any


class Process:
    """
    Starts all applications
    establish connections to discovery, queues
    """
    service_groups = []  # type: List[Dict[str, Any]]

    def __init__(self, discovery):
        self.discovery = discovery
        self.instances = []

    def create_context(self, service=None):
        return Context(self.discovery, service)

    def start(self):
        for SG in self.service_groups:
            sg = SG(self.discovery)
            sg.start()
            self.instances.append(sg)

    def stop(self):
        for sg in self.instances:
            sg.stop()



import logging
from copy import copy
from fan.contrib.aio.discovery import ZKDiscovery


logging.basicConfig(level=logging.DEBUG)

class SyncHelper:
    # url: viewset class
    # 'service_name': {'url', 'url_kwargs', 'args_names'}
    log = logging.getLogger('SyncHelper')

    def __init__(self, zk_config, chroot, config):
        self.d = ZKDiscovery(zk_config, chroot)
        self.config = config
        self.registered = {}

    async def on_start(self):
        await self.d.on_start()
        await self.register_all()

    async def register_all(self):
        for service in self.config['services']:
            zk_path, data = await self.register_endpoint(service)

    def gen_callback(self, path, data, service):
        async def updated(new_data):
            self.log.debug('New data: {!r}'.format(new_data))
            if new_data == data:
                return
            self.log.debug('Updated node: {} {} => {}'.format(path, data, new_data))
            self.d.unwatch(path, updated)
            await self.register_endpoint(service)
        return updated

    async def register_endpoint(self, service):
        service_original = service
        service = copy(service)
        name = ['/endpoints'] + service.pop('name').split('.')
        version = service.pop('version')
        path = name + [version]
        barrier_path = '/'.join(path + ['barrier'])

        zk_path, data = await self.d.register_raw(path, service, barrier_path)
        self.registered[zk_path] = (path, service, barrier_path)
        self.log.debug('Registered: {!r}'.format(zk_path))
        self.d.watch(zk_path, self.gen_callback(zk_path, data, service_original))
        return zk_path, data
